%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     
%              R O T A T I O N   S E A R C H
%
% This package contains the source code which implements the
% BnB rotation search algorithm and the nested 6 DoF registration 
% algorithm proposed in
% 
% A. Parra Bustos, T.-J. Chin, A. Eriksson, H. Li and D. Suter
% Fast Rotation Search with Stereographic Projections for 3D Registration
% IEEE Transactions on Pattern Analysis and Machine Intelligence (TPAMI)
%
% Copyright (c) 2016 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
% School of Computer Science, The University of Adelaide, Australia
% The Australian Center for Visual Technologies
% http://cs.adelaide.edu.au/~aparra
% Please acknowledge the authors by citing the above paper in any academic
% publications that have made use of this package or part of it.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% This demo contains sample experiments on 3 types of data/applications:
% 1. Rotation search on synthetic point correspondences.
% 2. Rotation search on real data.
% 3. Point cloud registration using nested BnB.


clear; clc;
close all;
addpath data src src_demo;

% Flag to choose between compile the code or use provided binary.
USE_BINARY=0;

if USE_BINARY
    % Test binary code
    fprintf('*****************************************************\n');
    fprintf('*     Running demo: using provided binary           *\n');
    fprintf('*****************************************************\n\n');
    
     try
        % Load point matches
        load( sprintf('ex1_data%d.mat', 100) );
        [~,~] = rot3(X',Y',2,20,20,1);
    catch exception
        fprintf('Failed execution of provided binary:\n');
        fprintf('%s\n\n',exception.message);
        fprintf('Compiling source code...\n');
        compile
     end
else
    % Compile GORE
    fprintf('*****************************************************\n');
    fprintf('*     Running demo: compiling code                  *\n');
    fprintf('*****************************************************\n\n');
    
    compile
end

results=struct();

%% Example 1: Synthesized data

fprintf('Example 1: rotation search for synthesized data.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

synth_data={};
dataSize=[100,250,500];
for i=1:length(dataSize)
    fprintf('Running GORE for %d point matches.\n',dataSize(i));
    
    % Load point matches
    load( sprintf('ex1_data%d.mat', dataSize(i)) );
    
    % Plot input points
    figure
    subplot(1,2,1);
    Xcloud =pointCloud(X);
    Ycloud =pointCloud(Y);
    Xcloud.Color = uint8(repmat([255 0 0],Xcloud.Count,1));
    Ycloud.Color = uint8(repmat([0 0 255],Ycloud.Count,1));
    pcshow(Xcloud,'MarkerSize',100);
    hold on
    pcshow(Ycloud,'MarkerSize',100);
    title('Input')
    drawnow

    lwbnd=0;
    g = struct();
   
    ml_flag=1;
    gap=0;
    th=1.5;
    tic
    [g.q, g.R] = rot3(X',Y',th,lwbnd,gap,ml_flag);
    g.time = toc;
   
    synth_data{end+1}=g;
    
    fprintf('Number of points............. %d\n', dataSize(i));
    fprintf('Time (s)..................... %f\n', g.time);
    fprintf('Quality...................... %d\n', g.q);
    fprintf('Output rotation matrix:');
    display(g.R);
    
    % Plot found rotation over all points.
    subplot(1,2,2);
    tform = affine3d([g.R' [0 0 0]'; 0 0 0 1 ]);
    X2cloud = pctransform(Xcloud,tform);
    pcshow(X2cloud,'MarkerSize',100);
    hold on
    pcshow(Ycloud,'MarkerSize',100);
    title('Optimal rotation')

end
results.synth_data=synth_data;

%% Example 2.
fprintf('Example 2: Rotation search in real data.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

% Load points
pts=load('ex2_data.mat');

X = pointCloud(pts.X);
Y = pointCloud(pts.Y);

gridStep=2;
X = pcdownsample(X,'gridAverage',gridStep);
Y = pcdownsample(Y,'gridAverage',gridStep);

%set colour
xsize=size(X.Location,1);
ysize=size(Y.Location,1);

X.Color = uint8(repmat([255 0 0],xsize,1));
Y.Color = uint8(repmat([0 0 255],ysize,1));

%initialisation
R=randrot();
tform = affine3d([R [0 0 0]'; 0 0 0 1 ]);
X = pctransform(X,tform);

% Plot input
figure;
subplot(1,2,1)
pcshow(X,'MarkerSize',100);
hold on
pcshow(Y,'MarkerSize',100);
drawnow

bnb=struct();
th=1.5;
lwbnd=0;
gap=0;
ml_flag=1; % To use matchlists

tic
[bnb.q, bnb.R] = rot3(X.Location',Y.Location',th,lwbnd,gap,ml_flag);
bnb.time=toc;

tform = affine3d([bnb.R' [0 0 0]'; 0 0 0 1 ]);
X2 = pctransform(X,tform);

subplot(1,2,2)
pcshow(X2,'MarkerSize',100);
hold on
pcshow(Y,'MarkerSize',100);


fprintf('Number of point.............. |X|=%d, |Y|=%d\n', xsize, ysize);
fprintf('Time (s)..................... %f\n', bnb.time);
fprintf('Optimal solution............. %d\n', bnb.q);
fprintf('Output rotation matrix:');
display(bnb.R);


%% Example 3
fprintf('Example 3: 6 DoF Point cloud registration.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

% Load points
pts=load('ex2_data.mat');

X = pointCloud(pts.X);
Y = pointCloud(pts.Y);

gridStep=7;
X = pcdownsample(X,'gridAverage',gridStep);
Y = pcdownsample(Y,'gridAverage',gridStep);

%set colour
xsize=size(X.Location,1);
ysize=size(Y.Location,1);

X.Color = uint8(repmat([255 0 0],xsize,1));
Y.Color = uint8(repmat([0 0 255],ysize,1));

%initialisation
%R=randrot();
%t=[.2 .2 .2];
init = load('exp3_init.mat');
R=init.R;
%t=init.t;
t=[15 15 15];

tform = affine3d([R [0 0 0]'; t 1 ]);
X = pctransform(X,tform);

% Plot input
figure;
subplot(1,2,1)
pcshow(X,'MarkerSize',100);
hold on
pcshow(Y,'MarkerSize',100);
drawnow

bnb=struct();
th=2.5;
lwbnd=0;
gap=0;
ml_flag=1; % To use matchlists
tDom=[-30 -30 -30 30 30 30]; %translation domain box
lm_flag=1; % To boost using a local method

tic
[bnb.q, bnb.T] = reg6(X.Location',Y.Location',th,...
                      tDom,lwbnd,gap,ml_flag,lm_flag);
bnb.time=toc;

tform = affine3d(bnb.T');
X2 = pctransform(X,tform);

subplot(1,2,2)
pcshow(X2,'MarkerSize',100);
hold on
pcshow(Y,'MarkerSize',100);

fprintf('Size of the input............ |X|=%d, |Y|=%d\n', xsize, ysize);
fprintf('Time (s)..................... %f\n', bnb.time);
fprintf('Optimal solution............. %d\n', bnb.q);
fprintf('Output transformation:');
display(bnb.T);
