/////////////////////////////////////////////////////////////////////////////
//
//              R O T A T I O N   S E A R C H
//
// This package contains the source code which implements the
// BnB rotation search algorithm and the nested 6 DoF registration
// algorithm proposed in
//
// A. Parra Bustos, T.-J. Chin, A. Eriksson, H. Li and D. Suter
// Fast Rotation Search with Stereographic Projections for 3D Registration
// IEEE Transactions on Pattern Analysis and Machine Intelligence (TPAMI)
//
// Copyright (c) 2016 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
// School of Computer Science, The University of Adelaide, Australia
// The Australian Center for Visual Technologies
// http://cs.adelaide.edu.au/~aparra
// Please acknowledge the authors by citing the above paper in any academic
// publications that have made use of this package or part of it.
//
/////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "reg_common.h"
#include "registration.h"
#include "mex.h"

using namespace reg;


// Input
#define X_IN     prhs[0]
#define Y_IN     prhs[1]
#define TH_IN    prhs[2]
#define TDOM_IN  prhs[3] // Translation search domain
#define LWBND_IN prhs[4]
#define GAP_IN   prhs[5] // BnB gap. Stopping threshold gap
#define ML_IN    prhs[6] // If 1 use matchlists
#define LM_IN    prhs[7] // If 1 use the local method
// Output
#define Q_OUT    plhs[0]  // Solution quality
#define T_OUT    plhs[1]  // Optimal transformation


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    
    // Check mex-arguments
    if (nlhs != 2)
    {
        mexErrMsgTxt("Invalid number of output variables. Expected 2 LHS vars.");
    }
    
    if (nrhs != 8)
    {
        mexErrMsgTxt("Invalid number of parameters. Expected 8 parameters.");
    }
    
    if (mxGetM(X_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }
    
    if (mxGetM(Y_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }
    
    if (!mxIsDouble(X_IN) || !mxIsDouble(Y_IN))
    {
        mexErrMsgTxt("Input argument must be of type double.");
    }
    
    // Retrieve input
    double *X = mxGetPr(X_IN);
    double *Y = mxGetPr(Y_IN);
    
    const double th = *mxGetPr(TH_IN);
    mxAssert(th>0,"");
    
    const int gap = (int)*mxGetPr(GAP_IN);
    mxAssert(gap>=0,"");
    
    const int lwbnd = (int)*mxGetPr(LWBND_IN);
    mxAssert(lwbnd>=0,"");
    
    const size_t xsize = mxGetN(X_IN);
    const size_t ysize = mxGetN(Y_IN);
    
    const bool ml = (bool)*mxGetPr(ML_IN);
    const bool lm = (bool)*mxGetPr(LM_IN);

    double *tDom = mxGetPr(TDOM_IN); // [a(1) a(1) a(1) b(1) b(2) b(3)]

    
    // Allocate output
    T_OUT = mxCreateDoubleMatrix(4, 4, mxREAL);
    
    Matrix3X matrixX;
    matrixX.n = xsize;
    matrixX.x = X;
    
    Matrix3X matrixY;
    matrixY.n = ysize;
    matrixY.x = Y;
    
    search::TranslationSearchSpaceRegion3DOF trSSR(tDom);
    
    Transform3 guessAndResult; //Identity
    
    size_t q;
    if(ml)
    {
        q = nestedbnb_search_6dof_mcirc_ml(matrixX, matrixY, th, gap, lwbnd, trSSR,
                                           guessAndResult, lm);
    }
    else
    {
        q = nestedbnb_search_6dof_mcirc(matrixX, matrixY, th, gap, lwbnd, trSSR,
                                        &reg::search::bnb_rsearch_3dof_mcirc,
                                        guessAndResult, lm);
    }
    
    //To do not free memory
    matrixX.x = matrixY.x = NULL;
    
    //Copy result
    memcpy(mxGetPr(T_OUT), guessAndResult.x, 16*sizeof(double) );
    Q_OUT = mxCreateDoubleScalar(q);
}

