/////////////////////////////////////////////////////////////////////////////
//
//              R O T A T I O N   S E A R C H
//
// This package contains the source code which implements the
// BnB rotation search algorithm and the nested 6 DoF registration
// algorithm proposed in
//
// A. Parra Bustos, T.-J. Chin, A. Eriksson, H. Li and D. Suter
// Fast Rotation Search with Stereographic Projections for 3D Registration
// IEEE Transactions on Pattern Analysis and Machine Intelligence (TPAMI)
//
// Copyright (c) 2016 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
// School of Computer Science, The University of Adelaide, Australia
// The Australian Center for Visual Technologies
// http://cs.adelaide.edu.au/~aparra
// Please acknowledge the authors by citing the above paper in any academic
// publications that have made use of this package or part of it.
//
/////////////////////////////////////////////////////////////////////////////


#include <iostream>
#include <ctime>
#include "reg_common.h"
#include "mex.h"
#include "registration.h"


using namespace reg;
using namespace reg::search;


// Input
#define X_IN     prhs[0]
#define Y_IN     prhs[1]
#define TH_IN    prhs[2]
#define LWBND_IN prhs[3]
#define GAP_IN   prhs[4] // BnB gap. Stopping threshold gap
#define ML_IN    prhs[5] // If 1 use matchlists
// Output
#define Q_OUT    plhs[0]  // Solution quality
#define R_OUT    plhs[1]  // Optimal rotation

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    // Check mex-arguments
    if (nlhs != 2)
    {
        mexErrMsgTxt("Invalid number of output variables. Expected 2 LHS vars.");
    }

    if (nrhs != 6)
    {
        mexErrMsgTxt("Invalid number of parameters. Expected 6 parameters.");
    }

    if (mxGetM(X_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }

    if (mxGetM(Y_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }
    
    if (!mxIsDouble(X_IN) || !mxIsDouble(Y_IN))
    {
        mexErrMsgTxt("Input argument must be of type double.");
    }
    

    // Retrieve input
    double *X = mxGetPr(X_IN);
    double *Y = mxGetPr(Y_IN);

    const double th = *mxGetPr(TH_IN);
    mxAssert(th>0,"");

    const int gap = (int)*mxGetPr(GAP_IN);
    mxAssert(gap>=0,"");

    const int lwbnd = (int)*mxGetPr(LWBND_IN);
    mxAssert(lwbnd>=0,"");

    const size_t xsize = mxGetN(X_IN);
    const size_t ysize = mxGetN(Y_IN);


    const bool ml = (bool)*mxGetPr(ML_IN);;

    // Allocate output

    R_OUT = mxCreateDoubleMatrix(3, 3, mxREAL);

    Matrix3X matrixX;
    matrixX.n = (int)xsize;
    matrixX.x = X;

    Matrix3X matrixY;
    matrixY.n = (int)ysize;
    matrixY.x = Y;
    
    const int buckets = MAX((int)ysize/10, 10);

    size_t q;
    AxisAngle resp;
    if (ml)
    {
        q = bnb_rsearch_3dof_mcirc_ml(matrixX, matrixY, th, gap, lwbnd, buckets, resp);
    }
    else
    {
        q = bnb_rsearch_3dof_mcirc(matrixX, matrixY, th, gap, lwbnd, buckets, resp);
    }

    //parse resp
    Matrix3 R;
    fromAxisAngle(R,resp);
    
    std::copy(R.getPr(), R.getPr()+9, mxGetPr(R_OUT));

    //To do not free memory
    matrixX.x = NULL;
    matrixY.x = NULL;

    Q_OUT = mxCreateDoubleScalar(q);
}
