/////////////////////////////////////////////////////////////////////////////
//
//              R O T A T I O N   S E A R C H
//
// This package contains the source code which implements the
// BnB rotation search algorithm and the nested 6 DoF registration
// algorithm proposed in
//
// A. Parra Bustos, T.-J. Chin, A. Eriksson, H. Li and D. Suter
// Fast Rotation Search with Stereographic Projections for 3D Registration
// IEEE Transactions on Pattern Analysis and Machine Intelligence (TPAMI)
//
// Copyright (c) 2016 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
// School of Computer Science, The University of Adelaide, Australia
// The Australian Center for Visual Technologies
// http://cs.adelaide.edu.au/~aparra
// Please acknowledge the authors by citing the above paper in any academic
// publications that have made use of this package or part of it.
//
/////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "registration.h"
#include "state_priority_hashtable.h"
#include "mcirc_indexation.h"
#include "search.h"

#define BRANCHING_FACTOR 8

using namespace reg;
using namespace reg::search;
typedef TranslationSearchSpaceRegion3DOF SSR;

inline
void translate(Matrix3X &TX, const Matrix3X &X, const Vector3 &t)
{
    const size_t n = X.cols();
    double *tx = TX.x;
    double *x = X.x;
    
    for (size_t i=0;i<n; i++)
    {
        tx[i*3]   = x[i*3]  +t.x;
        tx[i*3+1] = x[i*3+1]+t.y;
        tx[i*3+2] = x[i*3+2]+t.z;
    }
}


int reg::search::nestedbnb_search_6dof_mcirc(
        const Matrix3X &M, const Matrix3X &B,
        double th, int gap, int lwbnd,
        SSR &trRegion,
        int(*bnb_rsearch_3dof)(const Matrix3X&, const Matrix3X&,
                               double,int, int, int, AxisAngle&),
        Transform3 &guessAndResult,
        bool use_local_opt)
{
    int i, np;
    int state_lwbnd, upbnd, lqual;
    

    // Counters of bounds' evaluation calls
    int countUpbndEval, countLwbndEval;
    double dx, dy, dz;

    AxisAngle rsearchResult;
    Vector3 trCentre(0,0,0);
    Transform3 localGuessAndResult; //Identity

    const size_t msize = M.cols();
    const int buckets = MAX((int)msize/10,10);

    SearchState<SSR, int> *state;
    SSR **ssrArray;
    
    // Initialise
    countUpbndEval = countLwbndEval = 0;
    
    // Eval upper bound of guessAndResult transformation
    trCentre = ssrCentre(trRegion);
    Matrix3X TM(msize);
    translate(TM, M, trCentre);
    
    upbnd = bnb_rsearch_3dof(TM, B, th+ssrUncertainty(trRegion),
                              0, lwbnd, buckets, rsearchResult);
    countUpbndEval++;

    // No better solution than the known one in provided SSR
    mxAssert( upbnd>=lwbnd, "Bound error");
    if (upbnd-lwbnd <= gap)
    {
        return lwbnd;
    }

    ssrArray = new SSR*[BRANCHING_FACTOR];
    state = new SearchState<SSR>(trRegion, upbnd);

    StatePriorityHashtable<SSR, int, SearchState> table(buckets);
    table.push(state);

    int iter = 0;
    while (table.size())
    {
        iter++;
        state = table.pop();

        // Evaluate lower boud
        // Obtain lower bound calling the rotsearch with tr_c and 0 uncertainty trans.
        trCentre = ssrCentre(state->ssr);
        translate(TM, M, trCentre);
        state_lwbnd = bnb_rsearch_3dof(TM, B, th, 0, lwbnd, buckets, rsearchResult);

        countLwbndEval++;

        // Update solution
        if (state_lwbnd > lwbnd)
        {
            lwbnd = state_lwbnd;
            // update guessAndResult

            std::cout<<"Updating solution: curbest "<<lwbnd<<std::endl;
            
            Matrix3 R;
            fromAxisAngle(R,rsearchResult);
            Transform3 Rtfom(R);
            Transform3 Ttfom(trCentre);
            
            //guessAndResult = rsearchResult.toRotationMatrix() * Eigen::Translation3d(trCentre);

            guessAndResult = Rtfom * Ttfom;

            if(use_local_opt)
            {
                std::cout<<"calling local method..."<<std::endl;

                localGuessAndResult = guessAndResult;
                dx = .5*(state->ssr.bx-state->ssr.ax);
                dy = .5*(state->ssr.by-state->ssr.ay);
                dz = .5*(state->ssr.bz-state->ssr.az);
                SSR local_tr_box(-dx,-dy,-dz, dx,dy,dz);

                lqual = local_search_6dof(M, B, th, lwbnd, local_tr_box,
                                          bnb_rsearch_3dof, localGuessAndResult);
               
                std::cout<<"local method qual "<<lqual<<std::endl;
                if(lqual>lwbnd)
                {
                    lwbnd = lqual;
                    std::cout<< "curbest = "<<lwbnd<<std::endl;
                    guessAndResult = localGuessAndResult;
                }
            }

            //table.prune(curbest);
        }

        if(iter%10==0)
        {
            std::cout<< "iter "<< iter << " upbnd "<<state->bnd << " lwbnd "
                     << state_lwbnd<<" table size "<<table.size()<<std::endl;
        }

        // Stopping criterion
        if (state->bnd - lwbnd <= gap)
        {
            delete state;
            break;
        }

        // Branch
        np = reg::search::split(state->ssr, ssrArray);
        delete state;

        for(i=0; i<np; i++)
        {
            // Compute upper bound by calling rotsearch with tc and t uncertainty delta
            trCentre = ssrCentre( *(ssrArray[i]) );

            translate(TM, M, trCentre);

            upbnd = bnb_rsearch_3dof(TM,B,th+ssrUncertainty(*(ssrArray[i])),
                                     0, lwbnd, buckets, rsearchResult);

            if (upbnd >= lwbnd)
            {
                state = new SearchState<SSR>(*(ssrArray[i]), upbnd);
                table.push(state);
            }

            delete ssrArray[i];
        }
        countUpbndEval += np;

    } //bnb loop

    delete []ssrArray;
    return lwbnd;
}


int compare (const void * a, const void * b)
{
    return ( *(int*)a - *(int*)b );
}


//for lwbnd
int rsearch_ml( const reg::Matrix3X &M, const reg::Matrix3X &B,
                double th, int lwbnd,
                int buckets, int *matchList, int matchListSize,
                std::vector<bool> &matches, AxisAngle &result)
{
    mxAssert(th>=0 && matchListSize>0 && lwbnd>=0 && buckets>0,"invalid input");
    typedef RotationSearchSpaceRegion3DOFS8 RSSR;

    MCircIndexation<RSSR> dsi(M, B, th);
    dsi.sweep(matchList, matchListSize, matches);

    RSSR guessAndResult(-PI, -PI, -PI, PI, PI, PI);
    const int qual = reg::search::bnb_search_ml_table<RSSR, 8> (
                dsi, lwbnd, 0, buckets, guessAndResult);

    mxAssert(qual>=lwbnd && qual<=matchListSize, "invalid result");

    result = centre(guessAndResult);
    return qual;
}


//for upbnd
int rsearch_ml( const reg::Matrix3X &M, const reg::Matrix3X &B,
                double th,int lwbnd,
                int buckets, int *matchList, int matchListSize,
                std::vector<bool> &matches, int &sweepSize)
{
    mxAssert(th>=0 && matchListSize>0 && lwbnd>=0 && buckets>0, "invalid input");
    typedef RotationSearchSpaceRegion3DOFS8 RSSR;

    MCircIndexation<RSSR> dsi(M, B, th);
    sweepSize = dsi.sweep(matchList, matchListSize, matches);
    mxAssert(sweepSize>=0, "invalid sweep size");
    RSSR guessAndResult(-PI, -PI, -PI, PI, PI, PI);
    const int qual = reg::search::bnb_search_ml_table<RSSR, 8> (
                dsi, lwbnd, 0, buckets, guessAndResult);
 

    mxAssert(qual>=lwbnd && qual<=matchListSize, "wrong quality");
    return qual;
}


int rsearch( const reg::Matrix3X &M, const reg::Matrix3X &B,
             double th,int known_sol_qual, int buckets)
{
    mxAssert(th>=0 && known_sol_qual>=0 && buckets>0, "invalid input");
    typedef RotationSearchSpaceRegion3DOFS8 RSSR;

    MCircIndexation<RSSR> dsi(M, B, th);
    dsi.sweep();

    RSSR guessAndResult(-PI, -PI, -PI, PI, PI, PI);
    const int qual = reg::search::bnb_search_ml_table<RSSR, 8> (
                dsi, known_sol_qual, 0, buckets, guessAndResult);

    mxAssert(qual>=known_sol_qual, "invalid output");
    return qual;
}



int reg::search::nestedbnb_search_6dof_mcirc_ml(
        const Matrix3X &M, const Matrix3X &B,
        double th, int gap, int lwbnd,
        SSR &tr_region,
        Transform3 &guessAndResult,
        bool use_local_opt)
{
    int i,j,k, np;
    int state_lwbnd, upbnd, lqual;
    int countUpbndEval, countLwbndEval;
    double dx, dy, dz;
    int *matchList, *childMatchList;

    Vector3 trCentre(0,0,0);
    Transform3 localGuessAndResult; //Identity

    const size_t msize = M.cols();
    const int buckets = MAX((int)msize/10,10);

    StatePriorityHashtable<SSR, int, SearchStateML> table(buckets);


    // Initialise
    countUpbndEval = countLwbndEval= 0;


    // Eval upper bound of guessAndResult transformation
    matchList = new int[msize];
    for(i=0; i<msize; i++)
    {
        matchList[i]=i;
    }
    std::vector<bool> matchesMap(msize);
    int sweepSize;

    trCentre = ssrCentre(tr_region);
    Matrix3X TM(msize);
    translate(TM, M, trCentre);
    
    
//    size_t rsearch_ml( const reg::Matrix3X &M, const reg::Matrix3X &B,
//                      double th,int known_sol_qual,
//                      int buckets, int *matchList, size_t matchListSize,
//                      std::vector<bool> &matches, size_t &sweepSize)


    upbnd = rsearch_ml(TM, B, th+ssrUncertainty(tr_region),
                       lwbnd, buckets, matchList, msize,
                       matchesMap, sweepSize);

    mxAssert(upbnd==rsearch(TM, B, th+ssrUncertainty(tr_region),
                          lwbnd, buckets), "wrong upbnd");
    countUpbndEval++;

    mxAssert( upbnd>=lwbnd, "Bound error");
    if (upbnd-lwbnd <= gap)
    {
        delete matchList;
        return lwbnd;
    }

    //Create new matchlist
    childMatchList = new int[sweepSize];
    j=0;
    for(i=0; i<msize; i++)
    {
        if(matchesMap[i])
        {
            childMatchList[j++]=matchList[i];
        }
    }
    mxAssert(j>=upbnd, "wrong j");
    delete []matchList;

    SSR **ssrArray = new SSR*[BRANCHING_FACTOR];

    AxisAngle rsearchResult;
    SearchStateML<SSR, int> *state, *childState;
    state = new SearchStateML<SSR>(tr_region, upbnd, childMatchList, sweepSize);
    table.push(state);

    int iter = 0;
    while (table.size())
    {
        iter++;
        state = table.pop();

        // Evaluate lower boud
        // Obtain lower bound calling the rotsearch with tr_c and 0 uncertainty trans.
        trCentre = ssrCentre(state->ssr);
        translate(TM, M, trCentre);

        state_lwbnd = rsearch_ml(TM, B, th, lwbnd, buckets,
                           state->matchList, state->matchListSize,
                           matchesMap, rsearchResult);

        countLwbndEval++;

        // Update solution
        if (state_lwbnd > lwbnd)
        {
            lwbnd = state_lwbnd;
            // update guessAndResult

            std::cout<<"Updating solution: curbest "<<lwbnd<<std::endl;
            Matrix3 R;
            fromAxisAngle(R,rsearchResult);
            Transform3 Rtform(R);
            Transform3 Ttform(trCentre);
        
            //guessAndResult = rsearchResult.toRotationMatrix() * Eigen::Translation3d(trCentre);
            guessAndResult = Rtform*Ttform;
            
            if(use_local_opt)
            {
                // Call local method
                std::cout<<"calling local method..."<<std::endl;

                localGuessAndResult = guessAndResult;
                dx = .5*(state->ssr.bx-state->ssr.ax);
                dy = .5*(state->ssr.by-state->ssr.ay);
                dz = .5*(state->ssr.bz-state->ssr.az);
                SSR local_tr_box(-dx,-dy,-dz, dx,dy,dz);

                lqual = local_search_6dof(M, B, th, lwbnd, local_tr_box,
                                          bnb_rsearch_3dof_mcirc_ml,
                                          localGuessAndResult);

                std::cout<<"local method qual "<<lqual<<std::endl;
                if(lqual>lwbnd)
                {
                    lwbnd = lqual;
                    std::cout<< "curbest = "<<lwbnd<<std::endl;
                    guessAndResult = localGuessAndResult;
                }
            }

            table.prune(lwbnd);
        } // End update solution

        if(iter%10==0)
        {
            std::cout<< "iter "<< iter << " upbnd "<<state->bnd << " lwbnd "
                     << state_lwbnd<<" table size "<<table.size()<<std::endl;
        }

        // Stopping criterion
        if (state->bnd - lwbnd <= gap)
        {
            delete state;
            break;
        }

        // Branch
        np = reg::search::split(state->ssr, ssrArray);

        for(i=0; i<np; i++)
        {
            // Compute upper bound by calling rotsearch with tc and t uncertainty delta
            trCentre = ssrCentre( *(ssrArray[i]) );

            translate (TM, M, trCentre);

            upbnd = rsearch_ml(TM, B,th+ssrUncertainty(*(ssrArray[i])),
                               lwbnd, buckets,
                               state->matchList, state->matchListSize,
                               matchesMap, sweepSize);

            mxAssert(upbnd==rsearch(TM,B,th+ssrUncertainty(*(ssrArray[i])),lwbnd,buckets), "wrong upbnd");

            if ( upbnd >= lwbnd )
            {
                childMatchList = new int[sweepSize];

                k=0;
                for(j=0; j<state->matchListSize; j++)
                {
                    if(matchesMap[j])
                    {
                        childMatchList[k++]=state->matchList[j];
                    }
                }
                mxAssert(k>=upbnd, "wrong k");

                childState = new SearchStateML<SSR>(
                            *(ssrArray[i]),
                            upbnd, childMatchList, sweepSize);
                table.push(childState);
            }

            delete ssrArray[i];
        }
        delete state;
        countUpbndEval += np;

    } //bnb loop

    delete []ssrArray;
    return lwbnd;
}
